/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*      ___________________________________________________________________     */
/*      | H Y T E C   8 4 2 4  1 6 B i t  1 M  A D C  A s y n  D r i v e r |    */
/*           ---------------------------------------------------------------    */
/*                                                                              */
/*  Source file name      :-      drvHy8424ip.c                                 */
/*                                                                              */
/*  Initial creation date :-      03-Jan-2013                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec 8424 ADC EPICS support. It is based   */
/*  on Hy8417ipasyn driver code. It is osi complaint.		                    */
/*                                                                              */
/*  (C)2013 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    03/01/13    |  JC		  | 1.0     | Initial creation with new asyn model  */    
/*    15/02/13    |  JC		  | 2.0     | Added transit recorder support		*/    
/*    06/06/13    |  JC		  | 2.1     | Optimised the init functions			*/    
/*    17/01/14    |  JC		  | 2.2     | Added shell command to suit VxWorks	*/    
/*    			  |  		  |      	| for transient recorder mode. Also		*/    
/*    			  |  		  |      	| modified ISR to suit latest asyn.		*/    
/*    12/05/14    |  JC		  | 2.3     | Modified waveform data return with	*/    
/*    			  |  		  |      	| averaging (accumulated by firmware	*/    
/*                |           |         | and averaged by software) if so wished*/    
/*    06/10/14    |  JC		  | 2.4     | Modified to support 64bit Linux based	*/    
/*                |  M.Rivers |         | on Mark Rivers' driverTable model     */    
/*                |           |         | Note: need V2.6 of drvHyLinuxDriver.c */    
/*                |           |         | in ipac module to work with 64bit     */    
/*                |           |         | Linux		                            */    
/*                |           |         |    		                            */    
/********************************************************************************/

#define DRIVER_VERSION	204
//#define DEBUG

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8424Asyn.h"

static Hy8424Asyn *driverTable[MAX_CARDS];
static int numCards;

/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
 *				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
 *				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
 *				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz,
 *				  16 = 200khz, 17 = 500Khz, 18 = 1MHz.
 *     extClock = 0 = internal, 1 = extrenal,
 *		  range = 0 : +/-10V, 1 : +/-5V
 */    
Hy8424Asyn::Hy8424Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int clockRate, int extClock, int range)    
		: asynPortDriver(portName, MAX_ADC_CHANNELS, NUM_PARAMS, 
        asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask | asynDrvUserMask,
        asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok;
  	char lstr[7];
  	USHORT idmem;
    int ishytec;
    int ismodel;
	int ver;

	this->lastPointer = 0;

	/* sweep data channels definition */
    createParam(hy8424SweepDataString, 		asynParamInt32, 	&hy8424SweepData_);
    createParam(hy8424FloatSweepDataString, asynParamFloat64, 	&hy8424FloatSweepData_);

	/* normal data channels definition */
    createParam(hy8424DataString, 			asynParamInt32, 	&hy8424Data_);
    createParam(hy8424FloatDataString, 		asynParamFloat64, 	&hy8424FloatData_);

	/* controls definition */
    createParam(hy8424ScanPeriodString, 	asynParamFloat64, 	&hy8424ScanPeriod_);
    createParam(hy8424SetSampleString, 		asynParamInt32, 	&hy8424SetSample_);
    createParam(hy8424SetAverageString, 	asynParamInt32, 	&hy8424SetAverage_);
    createParam(hy8424SetModeString, 		asynParamInt32, 	&hy8424SetMode_);
    createParam(hy8424SetOffsetString, 		asynParamInt32, 	&hy8424SetOffset_);
    createParam(hy8424EnableString, 		asynParamInt32, 	&hy8424Enable_);
    createParam(hy8424SoftTriggerString, 	asynParamInt32, 	&hy8424SoftTrigger_);
    createParam(hy8424AutoReenableString, 	asynParamInt32, 	&hy8424AutoReenable_);
    createParam(hy8424ClearBufferString, 	asynParamInt32, 	&hy8424ClearBuffer_);
    createParam(hy8424SetClockRateString, 	asynParamInt32, 	&hy8424SetClockRate_);
    createParam(hy8424SetExtClockString, 	asynParamInt32, 	&hy8424SetExtClock_);

    createParam(hy8424RangeString, 			asynParamInt32, 	&hy8424Range_);
    createParam(hy8424FastADCString, 		asynParamInt32, 	&hy8424FastADC_);
    createParam(hy8424TriggerEdgeString, 	asynParamInt32, 	&hy8424TriggerEdge_);

    createParam(hy8424TRThresholdLowString, asynParamInt32, 	&hy8424TRThresholdLow_);
    createParam(hy8424TRThresholdHighString,asynParamInt32, 	&hy8424TRThresholdHigh_);
    createParam(hy8424TRPeakString, 		asynParamInt32, 	&hy8424TRPeak_);
    createParam(hy8424TRChannelString, 		asynParamInt32, 	&hy8424TRChannel_);
    createParam(hy8424TRBufferString, 		asynParamInt32, 	&hy8424TRBuffer_);
    createParam(hy8424TROutputEdgeString, 	asynParamInt32, 	&hy8424TROutputEdge_);
    createParam(hy8424TRInputEdgeString, 	asynParamInt32, 	&hy8424TRInputEdge_);
    createParam(hy8424SoftStopString, 		asynParamInt32, 	&hy8424SoftStop_);

    createParam(hy8424AverageOverflowString,asynParamInt32, 	&hy8424AverageOverflow_);
    createParam(hy8424OverflowString, 		asynParamInt32, 	&hy8424Overflow_);
    createParam(hy8424BufferCountString, 	asynParamInt32, 	&hy8424BufferCount_);
    createParam(hy8424MessageSentString, 	asynParamInt32, 	&hy8424MessageSent_);
    createParam(hy8424MessageFailedString, 	asynParamInt32, 	&hy8424MessageFailed_);
    createParam(hy8424TriggerStatusString,	asynParamInt32, 	&hy8424TriggerStatus_);
    createParam(hy8424FWVersionString, 		asynParamInt32, 	&hy8424FWVersion_);
    createParam(hy8424DriverVersionString,	asynParamInt32, 	&hy8424DriverVersion_);
    createParam(hy8424SupportString, 		asynParamInt32, 	&hy8424Support_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum		 = carrierNum;
    this->ipSlotNum			 = ipSlotNum;
    this->vectorNum          = vectorNum;

    this->lastLowPointer = 0;																					/* for periodical records */
	this->lastHighPointer = 0;
	this->firstTimeFlag = 0;																					//clear first time flag at the beginning

	setIntegerParam(hy8424SetClockRate_, clockRate);
	setIntegerParam(hy8424SetExtClock_, extClock);
	setIntegerParam(hy8424Range_, range);
	this->banksize = BANKSIZE;
	setIntegerParam(hy8424DriverVersion_, DRIVER_VERSION);

	/* Map Register address spaces  */
	this->adc = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (this->adc == NULL)
    {  
        /* can't register card register address space */
    	printf("configHy8424: Unable to register IO space\n");
        return;
    }
    
    /* map the memory address spaces  */
    this->convMemoryBase = (USHORT *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrMem);
    if (this->convMemoryBase == NULL)
    {  
        /* can't register card Memory address space */
     	printf("configHy8424: Unable to register memory space\n");
        return;
    }
     
    /* Compare the HY8424 prom header, manufacturer's ID and card model
     * number with the expected values. */
  	idmem =  *((USHORT*)this->adc->prom.vita4);
  	if((idmem == 0x5649) || (idmem == 0x4956));
  	{
  		lstr[0] = 0x56; lstr[1] = 0x49;
  	}
  	idmem =  *((USHORT*)this->adc->prom.vita4 + 1);
  	if((idmem == 0x5441) || (idmem == 0x4154));
  	{	
  		lstr[2] = 0x54; lstr[3] = 0x41;
  	}
  	idmem =  *((USHORT*)this->adc->prom.vita4 + 2);
  	if((idmem == 0x3420) || (idmem == 0x2034));
  	{	
  		lstr[4] = 0x20; lstr[5] = 0x34;
  	}
  	lstr[6] = 0;

    strok = memcmp(lstr, VITA4, sizeof(this->adc->prom.vita4)) == 0;
    ishytec = this->adc->prom.hytecIdHigh == HYTECID_HI  &&
                  this->adc->prom.hytecIdLow == HYTECID_LO;
    ismodel = this->adc->prom.model == PROM_MODEL;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("PROM: Contents at offset - %#08x\n", (int)&this->adc->prom);
    printf("PROM: Prom header string - '%6s'\n", lstr);
    printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
        (int)this->adc->prom.hytecIdHigh, (int)this->adc->prom.hytecIdLow, 
        ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
        (int)this->adc->prom.model, (int)this->adc->prom.revision);

	ver = (int)this->adc->prom.revision;
	ver = ver/4096*1000 + (ver%4096)/256*100 + ((ver%4096)%256)/16*10 + ((ver%4096)%256)%16;
	setIntegerParam(hy8424FWVersion_, ver);

    printf("PROM: Serial number      - %hu\n", (int)this->adc->prom.serialNumber);
    printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
        (int)this->adc->prom.driverIdHigh, (int)this->adc->prom.driverIdLow);
    printf("PROM: Device flags       - %#04hx\n", (int)this->adc->prom.flags);
    printf("PROM: Bytes used         - %#04hx (%hd)\n",
        (int)this->adc->prom.byteCount, (int)this->adc->prom.byteCount);
    printf("\n");

    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);
    if (!ismodel)
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n",
            PROM_MODEL);
	}

    /* clear asyn message sent counters */
    this->messagesSent = 0;
    this->messagesFailed = 0;

    /* set up interrupt, work queue and mutex stuff */
    this->adc->reg.vector = vectorNum;            /* writes vector register offset 0x08 */
    driverTable[numCards] = this;
    numCards++;
	if (ipmIntConnect(this->carrierNum, this->ipSlotNum, this->vectorNum, INTSRC, numCards-1) < 0) {
      printf("configHy8424: failure connecting to interrupt\n");
      return;
    }

    ipmIrqCmd(this->carrierNum, this->ipSlotNum, 0, ipac_irqEnable);

    this->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                             (MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8424intQueuedTask",
                           epicsThreadPriorityLow,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intQueuedTaskC,
                           (void *)this) == NULL)
	{	
      	printf("configHy8424: failure creating message queue.\n");
      	return;
	}
    
    this->intSweepQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8424intSweepQTask",
                           epicsThreadPriorityHigh,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intSweepQueuedTaskC,
                           (void *)this) == NULL)
	{
       	printf("Hy8424intSweepQueuedTask epicsThreadCreate failure\n");
      	return;
	}

	this->rebooting = 0;

   	epicsAtExit(rebootCallbackC, this);

}

/** Continue Mode, called directly or from iocsh */
/* 
 * Where: -
 *		average = number of readings to SUM for average
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 */
int Hy8424Asyn::Hy8424ContMode(int average, int fastADC)
{
	int mode = 0;

	setIntegerParam(hy8424SetMode_, mode);
	this->mode = mode;
	
	setIntegerParam(hy8424SetSample_, 0);
	setIntegerParam(hy8424SetOffset_, 0);
	if (average <= 0) average = 1;													//force it to be 1 if defined carelessly
	setIntegerParam(hy8424SetAverage_, average);
	setIntegerParam(hy8424FastADC_, fastADC);
	this->fastadc = fastADC;
	setIntegerParam(hy8424Enable_, 0);
	this->enable = 0;
	setIntegerParam(hy8424AutoReenable_, 0);
	setIntegerParam(hy8424BufferCount_, 0);

	// tell listeners the changes
	callParamCallbacks();

	// set up ADC 
	setupADC( INIT );
	return(asynSuccess);
}

/** Trigger Mode, called directly or from iocsh */
/* 
 * Where: -
 *      samples = number of samples (not applicable in gated modes),
 *		average = number of readings to SUM for average
 *       offset = Offset to start collecting after trigger
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 *	triggerEdge = 0 : rising edge; 1 : falling edge; 2 : both edge
 */
int Hy8424Asyn::Hy8424TrigMode(int sample, int average, double offset, int fastADC, int triggerEdge)
{
	int mode = 1;

	setIntegerParam(hy8424SetMode_, mode);
	this->mode = mode;
		
	setIntegerParam(hy8424SetSample_, sample);
	this->samples = sample;
	setIntegerParam(hy8424SetAverage_, average);
	setIntegerParam(hy8424SetOffset_, (int) offset);
	setIntegerParam(hy8424FastADC_, fastADC);
	this->fastadc = fastADC;
	setIntegerParam(hy8424TriggerEdge_, triggerEdge);

	setIntegerParam(hy8424TriggerStatus_, WAITING_TRIGGER);
	setIntegerParam(hy8424SoftTrigger_, 0);
	setIntegerParam(hy8424Enable_, 0);
	this->enable = 0;
	setIntegerParam(hy8424AutoReenable_, 0);
	setIntegerParam(hy8424BufferCount_, 0);

	// tell listeners the changes
	callParamCallbacks();

	//set up ADC
	setupADC( INIT );

	return(asynSuccess);
}

/** Gated Mode, called directly or from iocsh */
/* 
 * Where: -
 *		average = number of readings to SUM for average
 *       offset = Offset from the start (or end if negative) of the gated period to the ADC readings to return
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 */
int Hy8424Asyn::Hy8424GateMode(int average, double offset, int fastADC)
{
	int mode = 2;

	setIntegerParam(hy8424SetMode_, mode);
	this->mode = mode;
	
	setIntegerParam(hy8424SetAverage_, average);
	setIntegerParam(hy8424SetOffset_, (int) offset);
	setIntegerParam(hy8424FastADC_, fastADC);
	this->fastadc = fastADC;
	
	setIntegerParam(hy8424Enable_, 0);
	this->enable = 0;
	setIntegerParam(hy8424AutoReenable_, 0);
	setIntegerParam(hy8424BufferCount_, 0);

	// tell listeners the changes
	callParamCallbacks();

	// set up ADC
	setupADC( INIT );

	return(asynSuccess);
}

/** Transit recorder configuration command, called directly or from iocsh */
/* 
 * Where: -
 *		  sample = number of samples post trigger
 *         delay = delay time from the trigger to the start point of acquisition
 * thresholdlow0 = First threshold point for channel 0. In voltage unit
 *thresholdhigh0 = Second threshold point for channel 0. In voltage unit
 * thresholdlow1 = First threshold point for channel 1. In voltage unit
 *thresholdhigh1 = Second threshold point for channel 1. In voltage unit
 * thresholdlow2 = First threshold point for channel 2. In voltage unit
 *thresholdhigh2 = Second threshold point for channel 2. In voltage unit
 * thresholdlow3 = First threshold point for channel 3. In voltage unit
 *thresholdhigh3 = Second threshold point for channel 3. In voltage unit
 *    outputEdge = Start Out TTL level. 0=high, 1=low
 *    stopInEdge = Stop in TTL level: 0=high, 1=low
 */
int Hy8424Asyn::Hy8424SimpleTR(int sample, double delay, double thresholdlow0, double thresholdhigh0, double thresholdlow1, double thresholdhigh1,
						double thresholdlow2, double thresholdhigh2, double thresholdlow3, double thresholdhigh3, int outputEdge, int stopInEdge)
{
	int mode = 3, clockrate, i;
	
	setIntegerParam(hy8424SetMode_, mode);
	this->mode = mode;

	getIntegerParam(hy8424SetClockRate_, &clockrate);

	for(i=0; i<19; i++)
	{
		if((i%3) == 0) this->scanningRate[i] = (int) (1*pow(10,i/3));
		if(((i-1)%3) == 0) this->scanningRate[i] = (int) (2*pow(10,(i-1)/3));
		if(((i-2)%3) == 0) this->scanningRate[i] = (int) (5*pow(10,(i-2)/3));
	}

	setIntegerParam(hy8424SetSample_, sample);
	this->samples = sample;
	setIntegerParam(hy8424SetOffset_, (int) (delay*(double)this->scanningRate[clockrate]/TRDELAYDIVIDER));
	
	setIntegerParam(0, hy8424TRThresholdLow_, calcThreshold(thresholdlow0));
	setIntegerParam(0, hy8424TRThresholdHigh_, calcThreshold(thresholdhigh0));
	setIntegerParam(1, hy8424TRThresholdLow_, calcThreshold(thresholdlow1));
	setIntegerParam(1, hy8424TRThresholdHigh_, calcThreshold(thresholdhigh1));
	setIntegerParam(2, hy8424TRThresholdLow_, calcThreshold(thresholdlow2));
	setIntegerParam(2, hy8424TRThresholdHigh_, calcThreshold(thresholdhigh2));
	setIntegerParam(3, hy8424TRThresholdLow_, calcThreshold(thresholdlow3));
	setIntegerParam(3, hy8424TRThresholdHigh_, calcThreshold(thresholdhigh3));

	setTRLowThreshold(INIT, 0);
	setTRHighThreshold(INIT, 0);
	setTRLowThreshold(INIT, 1);
	setTRHighThreshold(INIT, 1);
	setTRLowThreshold(INIT, 2);
	setTRHighThreshold(INIT, 2);
	setTRLowThreshold(INIT, 3);
	setTRHighThreshold(INIT, 3);
	
	setIntegerParam(hy8424TROutputEdge_, outputEdge);
	setIntegerParam(hy8424TRInputEdge_, stopInEdge);

	setTROutEdge(INIT);
	setTRInEdge(INIT);
	
	// tell listeners the changes
	callParamCallbacks();
	
	// set up ADC
	setupADC( INIT );
	return(asynSuccess);
}


/** Transient recorder configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *		 sample = number of samples post trigger
 *    VTChannel = voltage trigger channel. 0~3
 * thresholdlow = First threshold point for VTChannel . In voltage unit
 *thresholdhigh = Second threshold point for VTChannel. In voltage unit
 *   outputEdge = Start Out TTL level. 0=high, 1=low
 *   stopInEdge = Stop in TTL level: 0=high, 1=low
 */
int Hy8424Asyn::Hy8424VoltageTR(int sample, int VTChannel, double thresholdlow, double thresholdhigh, int buffer, int outputEdge, int stopInEdge)
{
	int mode = 4;
	setIntegerParam(hy8424SetMode_, mode);
	this->mode = mode;

	setIntegerParam(hy8424SetSample_, sample);
	this->samples = sample;
	setIntegerParam(hy8424TRChannel_, VTChannel);
	setIntegerParam(hy8424TRBuffer_, buffer);
	setIntegerParam(hy8424TROutputEdge_, outputEdge);
	setIntegerParam(hy8424TRInputEdge_, stopInEdge);

	setIntegerParam(VTChannel, hy8424TRThresholdLow_, calcThreshold(thresholdlow));
	setIntegerParam(VTChannel, hy8424TRThresholdHigh_, calcThreshold(thresholdhigh));
	setTRLowThreshold(INIT, VTChannel);
	setTRHighThreshold(INIT, VTChannel);

	setTRChannel(INIT);
	setTRBuffer(INIT);
	setTROutEdge(INIT);
	setTRInEdge(INIT);

	callParamCallbacks();

	// set up ADC
	setupADC( INIT );
	return(asynSuccess);
}

/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
void Hy8424Asyn::report(FILE *fp, int details)
{
	int i, mode, range, extclock, clockrate, sample, average, offset, fwver, swver, fastadc, edge, TRlow, TRhigh;
	int VTTriggerChannel, VTBuffer, outedge, inedge;

    fprintf(fp, "Port: %s, carrier slot %d slot %d, base address=%p\n", 
            this->portName, this->carrierNum, this->ipSlotNum, (void *)this->adc);
    if (details >= 1) 
    {
		getIntegerParam(hy8424SetSample_, &sample);
		getIntegerParam(hy8424SetAverage_, &average);
		getIntegerParam(hy8424SetMode_, &mode);
		getIntegerParam(hy8424SetOffset_, &offset);
		getIntegerParam(hy8424SetClockRate_, &clockrate);
		getIntegerParam(hy8424SetExtClock_, &extclock);
		getIntegerParam(hy8424FWVersion_, &fwver);
		getIntegerParam(hy8424FastADC_, &fastadc);
		getIntegerParam(hy8424Range_, &range);
		getIntegerParam(hy8424DriverVersion_, &swver);
		getIntegerParam(hy8424TriggerEdge_, &edge);
		
        fprintf(fp, "  CSR=%x\n", this->adc->reg.status);
        fprintf(fp, "  EXTCSR=%x\n", this->adc->reg.extStatus);
        fprintf(fp, "  CSRTR=%x\n", this->adc->reg.trstatus);
        fprintf(fp, "  INT Vector=%x\n", this->vectorNum);
        fprintf(fp, "  Sample=%d\n", sample);
        fprintf(fp, "  Number of conversions=%d\n", this->adc->reg.numConversions[0]+(this->adc->reg.numConversions[1]>>16));
        fprintf(fp, "  Average=%d\n", average);
        fprintf(fp, "  Mode=%s\n", mode==0?"Normal":(mode==1?"Trigger":(mode==2?"Gated":(mode==3?"Simple TR":"VT TR"))));
        if(mode != hy8424TRSimple) fprintf(fp, "  Offset=%d\n", offset);
        fprintf(fp, "  Clock Source=%s\n", extclock==0?"Internal":"External");
        fprintf(fp, "  Clock Rate=%d\n", clockrate);
        fprintf(fp, "  Range=%s\n", range==0?"+/-10V":"+/-5V");
        fprintf(fp, "  Fast ADC=%s\n", fastadc==1?"Yes":"No");
        fprintf(fp, "  External trigger edge: %s\n", edge?"Falling":"Rising");
        fprintf(fp, "  Firmware Version=%d\n", fwver);
        fprintf(fp, "  Driver Version=%d\n", swver);
        fprintf(fp, "  messages sent OK=%d; send failed (queue full)=%d\n",
                this->messagesSent, this->messagesFailed);
        fprintf(fp, "  Counter=%d\n", this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16));
        
        if((mode == hy8424TRSimple) || (mode == hy8424TRVT))
		{
			getIntegerParam(hy8424TRChannel_, &VTTriggerChannel);
			getIntegerParam(hy8424TRBuffer_, &VTBuffer);
			getIntegerParam(hy8424TROutputEdge_, &outedge);
			getIntegerParam(hy8424TRInputEdge_, &inedge);
	
			if(mode == hy8424TRSimple) fprintf(fp, "  delay %d samples.\n", offset);
			if(mode == hy8424TRVT) 
			{
				fprintf(fp, "  transient recorder voltage trigger channel=%d.\n", VTTriggerChannel);
				fprintf(fp, "  transient recorder voltage trigger circular buffer size=0x%x.\n", VTBuffer);
			}
			fprintf(fp, "  transient recorder output trigger edge=%s.\n", (outedge == 0)?"High to trigger":"Low to trigger");
			fprintf(fp, "  transient recorder stop input edge=%s.\n", (inedge == 0)?"High to stop":"Low to stop");
			for(i=0; i<4; i++)
			{
				getIntegerParam(i, hy8424TRThresholdLow_, &TRlow);
				getIntegerParam(i, hy8424TRThresholdHigh_, &TRhigh);
				fprintf(fp, "  Channel%d low threshold raw = %d.\n", i, TRlow);
				fprintf(fp, "  Channel%d high threshold raw = %d.\n", i, TRhigh);
			}
		}       
    }
    asynPortDriver::report(fp, details);
}


/*****         asynInt32 methods       ******/
asynStatus Hy8424Asyn::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;

    this->getAddress(pasynUser, &channel);
	
/*	printf("\n*** hy8424_int32Write *** command = %d (%s), channel = %d val = %i\n\n",
				command, hy8424_commands[command].commandString, channel, value);
*/	
    status = setIntegerParam(function, value);

	/* decide what to do depending on which command we received */
	if (function == hy8424SetMode_)
	{
		this->mode = mode;
		setupADC( RUNTIME );
	}
	else if (function == hy8424Enable_)
		setEnableDisable( value );
	else if (function == hy8424SetAverage_)
        setAverage( RUNTIME );
	else if (function == hy8424SetSample_)
	{
		this->samples = value;
        setSamples( RUNTIME );
	}
	else if (function == hy8424SoftTrigger_)
        softwareTrigger();
	else if (function == hy8424SetClockRate_)
        setClockRate( RUNTIME );
	else if (function == hy8424SetExtClock_)
        setExtClock( RUNTIME );
	else if (function == hy8424Range_)
        setRange( value );
	else if (function == hy8424TRChannel_)
	{
		setIntegerParam(hy8424TRChannel_, value);
        setTRChannel( RUNTIME );
	}
	else if (function == hy8424TRBuffer_)
	{
		setIntegerParam(hy8424TRBuffer_, value);
        setTRBuffer( RUNTIME );
	}
	else if (function == hy8424TROutputEdge_)
	{
		setIntegerParam(hy8424TROutputEdge_, value);
        setTROutEdge( RUNTIME );
	}
	else if (function == hy8424TRInputEdge_)
	{
		setIntegerParam(hy8424TRInputEdge_, value);
        setTRInEdge( RUNTIME );
	}
	else if (function == hy8424SoftStop_)
        softwareStop();

	// tell listeners the changes
  	callParamCallbacks();

	return( status );
}

asynStatus Hy8424Asyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel, localValue=0;
    int pointer, numToAverage = 0, startPoint = 0;
    double dValue = 0.0;
	int function = pasynUser->reason;
	int mode, average, offset, triggerState;
    if (this->rebooting) epicsThreadSuspendSelf();
   	
	getIntegerParam(hy8424SetAverage_, &average);
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetOffset_, &offset);
	getIntegerParam(hy8424TriggerStatus_, &triggerState);
	
	this->getAddress(pasynUser, &channel);
		
    if(function == hy8424Data_)
	{
    	pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);

		switch(mode)
        {
        	case hy8424Continuous:
               	if (pointer > 0) 
                   	startPoint = pointer - SAMPLEWORDS;
               	else
                   	startPoint = this->banksize - SAMPLEWORDS;
               	break;
           	case hy8424Triggered:
			case hy8424Gated:
				if (pointer == 0) pointer = this->lastPointer; 						/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
               	if (pointer > 0) 
                   	startPoint = pointer - SAMPLEWORDS;
               	else
                   	startPoint = 0;
               	break;
           	default:
               	status = asynError;
           		return( status );
           		break;
        }

		numToAverage = average;
        dValue = (double) ((USHORT) ((this->convMemoryBase + channel * this->banksize)[startPoint]) +
        	(USHORT) ((this->convMemoryBase + channel * this->banksize)[startPoint+1]) * 65536 );
       	if (numToAverage != 0)
            localValue = (epicsInt32) (dValue / numToAverage);                  			/* check deviding 0 for safety */
        else
            localValue = (epicsInt32) dValue; 
        *value = localValue;

/*
if(channel == 0)
	printf("csr=0x%x xcsr=0x%x pt=%d stpt=%d value=%d  \n",
		this->adc->reg.status, this->adc->reg.extStatus, pointer, 
		startPoint, *value);                              
*/
	}
	else if(function == hy8424MessageSent_)
        *value = this->messagesSent;											// Good interrupts
	else if(function == hy8424MessageFailed_)
        *value = this->messagesFailed;											// Lost interrupts
	else if(function == hy8424BufferCount_)
        *value = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);	// memory pointer
	else if(function == hy8424Support_)
        *value = DRIVER_SUPPORT;                                                /* driver support */
	else if(function == hy8424TriggerStatus_)
	{
        if (mode == hy8424Continuous) 
			return asynError;                                                 /* continuous mode doesn't support this */
        else if (mode == hy8424Triggered)
            localValue = this->adc->reg.extStatus & EXTCSR_TEN;
        else if (mode == hy8424Gated)
            localValue =  (~this->adc->reg.status & CSR_MIS);

        if (localValue > 0)
		{                                                     					/* great than 0 means triggered */
            *value = TRIGGERRED_COLLECTING;
			setIntegerParam(hy8424TriggerStatus_, TRIGGERRED_COLLECTING);
		}
        else if (localValue == 0)                                               /* return whatever it is either waiting for trigger or finished or triigered by software */
            *value = triggerState;
    }
	else if(function == hy8424TRThresholdLow_)
	{
		*value = this->adc->reg.tr[channel].ptrLow[0] + (this->adc->reg.tr[channel].ptrLow[1] << 16);
	}
	else if(function == hy8424TRThresholdHigh_)
	{
		*value = this->adc->reg.tr[channel].ptrHigh[0] + (this->adc->reg.tr[channel].ptrHigh[1] << 16);
	}
	else if(function == hy8424TRPeak_)
        *value = this->adc->reg.tr[channel].peak;								/* for peak value */
	else
		asynPortDriver::readInt32(pasynUser, value);
	
  	callParamCallbacks();
	return( status );
}

asynStatus Hy8424Asyn::getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = 65535;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}

/* asynInt32Array methods */
asynStatus Hy8424Asyn::readInt32Array(asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, startPoint = 0, pointer, uppointer, numToReturn = 0;
	int function = pasynUser->reason;
	int mode, offset, trchannel, buffer, samples, average;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetOffset_, &offset);
	getIntegerParam(hy8424TRChannel_, &trchannel);
	getIntegerParam(hy8424TRBuffer_, &buffer);
	getIntegerParam(hy8424SetSample_, &samples);
	getIntegerParam(hy8424SetAverage_, &average);

	this->getAddress(pasynUser, &channel);

    if(function == hy8424Data_)
	{
    	pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);
        switch(mode)
        {
            case hy8424Continuous:
                if ((size_t) pointer/SAMPLEWORDS >= nelements) 
                    startPoint = (pointer - nelements*SAMPLEWORDS);
                else
                    startPoint = this->banksize - (nelements*SAMPLEWORDS - pointer);
                numToReturn = nelements;
                break;
            case hy8424Triggered:
                if (pointer == 0) pointer = this->lastPointer;  						//* get pointer saved by the interrupt but only when the pointer is reset by the enable function 
				status = setIntegerParam(hy8424Overflow_, 0);        					/* no overflow first */
				if (pointer / SAMPLEWORDS < (int)nelements)									//total samples less than required
				{
					if (pointer / SAMPLEWORDS >((int)nelements - offset))					//but still greater than really required
					{
						startPoint = pointer - (nelements - offset) * SAMPLEWORDS;
						numToReturn = nelements - offset;
					}
					else
					{
						startPoint = 0;													//less than really required
						numToReturn = startPoint / SAMPLEWORDS;
						status = setIntegerParam(hy8424Overflow_, 1);        			/* overflow, samples collected is less than offset */
					}
				}
				else
				{
					startPoint = pointer - (nelements - offset) * SAMPLEWORDS;			//total number greater than required
					numToReturn = nelements - offset;
				}
                break;
            case hy8424Gated:
                if (pointer == 0) pointer = this->lastPointer;  						//* get pointer saved by the interrupt but only when the pointer is reset by the enable function 
                if (offset >= 0)
                {
					status = setIntegerParam(hy8424Overflow_, 0);        				/* no overflow first */
					if (pointer / SAMPLEWORDS < (int)nelements)								//total samples less than required
					{
						if (pointer / SAMPLEWORDS >((int)nelements - offset))				//but still greater than really required
						{
							startPoint = pointer - (nelements - offset) * SAMPLEWORDS;
							numToReturn = nelements - offset;
						}
						else
						{
							startPoint = 0;												//less than really required
							numToReturn = startPoint / SAMPLEWORDS;
							status = setIntegerParam(hy8424Overflow_, 1);        		/* overflow, samples collected is less than offset */
						}
					}
					else
					{
						startPoint = pointer - (nelements - offset) * SAMPLEWORDS;		//total number greater than required
						numToReturn = nelements - offset;
					}
				}
                else
                {																		//negative offset
					status = setIntegerParam(hy8424Overflow_, 0);        				// no overflow first
					if (pointer / SAMPLEWORDS < ((int)nelements - offset))					//total samples less than required + offset
					{
						if (pointer / SAMPLEWORDS < (int)nelements)							//and also less than required, no choice but take everything
						{
							startPoint = 0;
							numToReturn = pointer / SAMPLEWORDS;
							status = setIntegerParam(hy8424Overflow_, 1);        		//overflow, samples collected is less than offset
						}
						else
						{																//yet greater than reuired, take as closer as possible to the really required
							startPoint = 0;												
							numToReturn = nelements;
							status = setIntegerParam(hy8424Overflow_, 1);        		//overflow, samples collected is less than offset
						}
					}
					else
					{
						startPoint = pointer - (nelements - offset) * SAMPLEWORDS;		//total number greater than required
						numToReturn = nelements;
					}
				}
                break;
            case hy8424TRSimple:
				startPoint = 0;
				numToReturn = samples;
                break;
			case hy8424TRVT:
				break;
			default:
                status = asynError;
                break;
        }

		if(mode != hy8424TRVT)
		{
			for(i=0; i<(int)nelements; i++)
    	    {
    	        if (i < numToReturn)
    	        {
					if (mode == hy8424TRSimple)
					{
						*(data + i) = (epicsInt32)(this->convMemoryBase + channel * this->banksize)[startPoint];
						startPoint ++;
					}
					else
					{
						*(data + i) = ((epicsInt32)(this->convMemoryBase + channel * this->banksize)[startPoint] +
							(this->convMemoryBase + channel * this->banksize)[startPoint + 1] * 65536) / average;
						startPoint += 2;
					}
    	        }
    	        else
    	            *(data + i) = 0;
    	        
    	        if(startPoint >= this->banksize)
    	        	startPoint = 0;                          							// if pointer reaches the bottom of the memory, reset it 
    	    }
	        *nactual = (int)nelements >= numToReturn ? numToReturn : nelements;         /* if number to return is greater or equal number of elements, NORD = number of elements */
		}else																			//* transient recorder voltage trigger mode 
		{
			if((pointer < buffer) && (pointer > 0))
			{
				startPoint = pointer;
				numToReturn = buffer;
			}
			else if(pointer > buffer)
			{
				startPoint = 0;
				numToReturn = pointer;
			}
			else		//pointer == 0
			{
				startPoint = 0;
				numToReturn = buffer + samples;
			}
			
			for(i=0; i<(int)nelements; i++)
		    {
		        if (i < numToReturn)
		        {
	            	*(data + i) = (this->convMemoryBase + channel * this->banksize)[startPoint];
	            	startPoint ++;
		        }
		        else
				{
					*(data + i) = 0;
					continue;
				}
		        
		        if((pointer < buffer) && (pointer > 0))
		        {
		        	if(startPoint >= numToReturn)
		        		startPoint = 0;                          										/* if pointer reaches the bottom of the memory, reset it */
		        }
		    }
	        *nactual = (int)nelements >= numToReturn ? numToReturn : nelements;              			/* if number to return is greater or equal number of elements, NORD = number of elements */
		}
    }
    else if(function == hy8424SweepData_)
	{    	   	
		switch (mode)
		{
			case hy8424Continuous:
				pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1] << 16);
				if (pointer > 0) 
					pointer = pointer - SAMPLEWORDS;
				else
					pointer = this->banksize - SAMPLEWORDS;
				for (i = 0; i<MAX_ADC_CHANNELS; i++)
				{
					*(data + i) = ((epicsInt32)(this->convMemoryBase + i * this->banksize)[pointer]) / average;
				}

				*nactual = MAX_ADC_CHANNELS;
				break;
			case hy8424Triggered:
			case hy8424Gated:
				pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1] << 16);
				if (pointer == 0) pointer = this->lastPointer;  										/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
				if (pointer > 0) pointer = pointer - SAMPLEWORDS;
				for(i=0; i<MAX_ADC_CHANNELS; i++)
        		{
      				*(data + i) = ((epicsInt32) (this->convMemoryBase + i * this->banksize)[pointer]) / average;
        		}
        
        		*nactual = MAX_ADC_CHANNELS;
				break;
			case hy8424TRSimple:
				startPoint = this->adc->reg.tr[channel].ptrLow[0] + (this->adc->reg.tr[channel].ptrLow[1] << 16);
				uppointer = this->adc->reg.tr[channel].ptrHigh[0] + (this->adc->reg.tr[channel].ptrHigh[1] << 16);

				if((startPoint == 0) && (uppointer == 0))												// if both are 0, first try the ones from interrupt
				{
					startPoint = this->lastLowPointer;
					uppointer = this->lastHighPointer;
				}
				
				if((startPoint == 0) && (uppointer == 0))
				{
					for(i=0; i<(int)nelements; i++)
						*(data + i) = 0;																/* if all trails the pointers are still 0, nothing to present */
					*nactual = (int)nelements;
				}
				else
				{
					if(uppointer == 0)																	/* if upper pointer is 0, check if it is full */
						numToReturn = samples;															/* if ADC memory is full, give it maximum */
					else
						numToReturn = uppointer;														/* or to the pointer */

					for(i=0; i<(int)nelements; i++)
					{
						if (i < numToReturn)
						{
							*(data + i) = (epicsInt32) (this->convMemoryBase + channel * this->banksize)[startPoint];
							startPoint++;
						}
						else
							*(data + i) = 0;
            
						if(startPoint >= this->banksize)
							startPoint = 0;                          									/* if pointer reaches the bottom of the memory, reset it */
					}
					*nactual = (int)nelements >= numToReturn ? numToReturn : nelements;              	/* if number to return is greater or equal number of elements, NORD = number of elements */
				}
				break;
			case hy8424TRVT:
				startPoint = this->adc->reg.tr[trchannel].ptrLow[0] + (this->adc->reg.tr[trchannel].ptrLow[1] << 16);
				uppointer = this->adc->reg.tr[trchannel].ptrHigh[0] + (this->adc->reg.tr[trchannel].ptrHigh[1] << 16);
				
				if((startPoint == 0) && (uppointer == 0))
				{
					for(i=0; i<(int)nelements; i++)
						*(data + i) = 0;
					*nactual = nelements;
				}
				else 
				{
					startPoint = buffer;
					if(uppointer > 0)
						numToReturn = uppointer - buffer;
					else
						numToReturn = samples;
		
					for(i=0; i<(int)nelements; i++)
		    	    {
        			    if (i < numToReturn)
		    	        {
   			        		*(data + i) = (epicsInt32) (this->convMemoryBase + channel * this->banksize)[startPoint];
            	    		startPoint++;
		    	        }
        			    else
            	    		*(data + i) = 0;
            		}
			        *nactual = (int)nelements >= numToReturn ? numToReturn : nelements;              /* if number to return is greater or equal number of elements, NORD = number of elements */
				}
				break;
			default:
				break;
		}
	}
	status = asynSuccess;
	return( status );
}

asynStatus Hy8424Asyn::writeFloat64(asynUser *pasynUser, epicsFloat64 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason, mode, rate;
	
    if (this->rebooting) epicsThreadSuspendSelf();
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetClockRate_, &rate);

	this->getAddress(pasynUser, &channel);
	
	/* decide what to do depending on which command we received */
	if (function == hy8424TRThresholdLow_)
	{
		setIntegerParam(channel, hy8424TRThresholdLow_, calcThreshold(value));
        setTRLowThreshold( RUNTIME, channel );	
	}
	else if (function == hy8424TRThresholdHigh_)
	{
		setIntegerParam(channel, hy8424TRThresholdHigh_, calcThreshold(value));
        setTRHighThreshold( RUNTIME, channel );
	}
	else if (function == hy8424SetOffset_)
	{
		if(mode != hy8424TRSimple)
		{
			setIntegerParam(hy8424SetOffset_, (int) value);
        	setOffset( RUNTIME );
		}
		else
		{
			setIntegerParam(hy8424SetOffset_, (int) (value*this->scanningRate[rate]/TRDELAYDIVIDER));
			setOffset( RUNTIME );
		}
	}

  	callParamCallbacks();
	return( status );
}


asynStatus Hy8424Asyn::readFloat64(asynUser *pasynUser, epicsFloat64 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
    int pointer, startPoint = 0, numToAverage = 0;
    double dValue = 0.0, power = 0.0;
	int mode, average, offset, period = 0, rate, readvalue;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetAverage_, &average);
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetOffset_, &offset);
	getIntegerParam(hy8424SetClockRate_, &rate);
	

	this->getAddress(pasynUser, &channel);
	
    if((function == hy8424Data_) || (function == hy8424FloatData_))
	{
    	pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);
    	
		switch(mode)
        {
		case hy8424Continuous:
			if (pointer > 0)
				startPoint = pointer - SAMPLEWORDS;
			else
				startPoint = this->banksize - SAMPLEWORDS;
			break;
		case hy8424Triggered:
		case hy8424Gated:
			if (pointer == 0) pointer = this->lastPointer; 						/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
			if (pointer > 0)
				startPoint = pointer - SAMPLEWORDS;
			else
				startPoint = 0;
			break;
        default:
            status = asynError;
           	return( status );
           	break;
        }

		numToAverage = average;
        dValue = (double) ((USHORT) ((this->convMemoryBase + channel * this->banksize)[startPoint]) +
        	(USHORT) ((this->convMemoryBase + channel * this->banksize)[startPoint+1]) * 65536 );

        if (numToAverage != 0)                                                  		/* safety check */
            *value = toVoltage(dValue/numToAverage);
        else
            *value = dValue;
    }
    else if((function == hy8424SweepData_) || (function == hy8424FloatSweepData_))      /* for FastSweep record, the most recent acquired data is used */
	{
		pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1] << 16);
        if ((mode == hy8424Triggered) || (mode == hy8424Gated))
        {
            if (pointer == 0) pointer = this->lastPointer;  							/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
        }
		if (pointer > 0)
			pointer = pointer - SAMPLEWORDS;
		else
		{
			if (mode == hy8424Continuous)
				pointer = this->banksize - SAMPLEWORDS;
			else
				pointer = 0;
		}

        dValue = (double) ((this->convMemoryBase + channel * this->banksize)[pointer] / average);
    	
        *value = dValue;
	}
	// sort out scan_period
	else if (function == hy8424ScanPeriod_)
	{
		getIntegerParam(hy8424SetClockRate_, &rate);
        period = (int) rate % 3;
        if((period == 0) || (period == 2))
            period = period * 2 + 1;
        else
            period = 2;
        power = pow((double) 10, (double) ((int) (rate / 3)));
        dValue = (double) 1 / (power * period);                             			/* this convert the frequency to second */
        *value = dValue;
        setDoubleParam(hy8424ScanPeriod_, dValue);
	}
	else if(function == hy8424TRPeak_)
        *value = toVoltage((double) this->adc->reg.tr[channel].peak);	 				// for peak voltage value 
	if (function == hy8424TRThresholdLow_)
	{
		getIntegerParam(channel, hy8424TRThresholdLow_, &readvalue);
		*value = calcThresholdRaw(readvalue);	
	}
	else if (function == hy8424TRThresholdHigh_)
	{
		getIntegerParam(channel, hy8424TRThresholdHigh_, &readvalue);
		*value = calcThresholdRaw(readvalue);	
	}
	else if (function == hy8424SetOffset_)
	{
		if(mode != hy8424TRSimple)
		{
			getIntegerParam(hy8424SetOffset_, &readvalue);
			*value = (double) readvalue;
		}
		else
		{
			getIntegerParam(hy8424SetOffset_, &readvalue);
			*value = (double)readvalue * TRDELAYDIVIDER / this->scanningRate[rate];
		}
	}
	else
		asynPortDriver::readFloat64(pasynUser, value);
	
  	callParamCallbacks();
	return( status );
}

/* asynFloat64Array methods */
asynStatus Hy8424Asyn::readFloat64Array(asynUser *pasynUser, epicsFloat64 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, pointer, startPoint = 0, uppointer, numToReturn = 0;
	int function = pasynUser->reason;
	int mode, offset, trchannel, buffer, samples, average;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetOffset_, &offset);
	getIntegerParam(hy8424TRChannel_, &trchannel);
	getIntegerParam(hy8424TRBuffer_, &buffer);
	getIntegerParam(hy8424SetSample_, &samples);
	getIntegerParam(hy8424SetAverage_, &average);

	this->getAddress(pasynUser, &channel);
	
    if((function == hy8424Data_) || (function == hy8424FloatData_))
    {
    	pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);
		switch (mode)
		{
		case hy8424Continuous:
			if ((size_t)pointer / SAMPLEWORDS >= nelements)
				startPoint = (pointer - nelements*SAMPLEWORDS);
			else
				startPoint = this->banksize - (nelements*SAMPLEWORDS - pointer);
			numToReturn = nelements;
			break;
		case hy8424Triggered:
			if (pointer == 0) pointer = this->lastPointer;  						//* get pointer saved by the interrupt but only when the pointer is reset by the enable function 
			status = setIntegerParam(hy8424Overflow_, 0);        					/* no overflow first */
			if (pointer / SAMPLEWORDS < (int)nelements)									//total samples less than required
			{
				if (pointer / SAMPLEWORDS >((int)nelements - offset))					//but still greater than really required
				{
					startPoint = pointer - (nelements - offset) * SAMPLEWORDS;
					numToReturn = nelements - offset;
				}
				else
				{
					startPoint = 0;													//less than really required
					numToReturn = startPoint / SAMPLEWORDS;
					status = setIntegerParam(hy8424Overflow_, 1);        			/* overflow, samples collected is less than offset */
				}
			}
			else
			{
				startPoint = pointer - (nelements - offset) * SAMPLEWORDS;			//total number greater than required
				numToReturn = nelements - offset;
			}
			break;
		case hy8424Gated:
			if (pointer == 0) pointer = this->lastPointer;  						//* get pointer saved by the interrupt but only when the pointer is reset by the enable function 
			if (offset >= 0)
			{
				status = setIntegerParam(hy8424Overflow_, 0);        				/* no overflow first */
				if (pointer / SAMPLEWORDS < (int)nelements)								//total samples less than required
				{
					if (pointer / SAMPLEWORDS >((int)nelements - offset))				//but still greater than really required
					{
						startPoint = pointer - (nelements - offset) * SAMPLEWORDS;
						numToReturn = nelements - offset;
					}
					else
					{
						startPoint = 0;												//less than really required
						numToReturn = startPoint / SAMPLEWORDS;
						status = setIntegerParam(hy8424Overflow_, 1);        		/* overflow, samples collected is less than offset */
					}
				}
				else
				{
					startPoint = pointer - (nelements - offset) * SAMPLEWORDS;		//total number greater than required
					numToReturn = nelements - offset;
				}
			}
			else
			{																		//negative offset
				status = setIntegerParam(hy8424Overflow_, 0);        				// no overflow first
				if (pointer / SAMPLEWORDS < ((int)nelements - offset))					//total samples less than required + offset
				{
					if (pointer / SAMPLEWORDS < (int)nelements)							//and also less than required, no choice but take everything
					{
						startPoint = 0;
						numToReturn = pointer / SAMPLEWORDS;
						status = setIntegerParam(hy8424Overflow_, 1);        		//overflow, samples collected is less than offset
					}
					else
					{																//yet greater than reuired, take as closer as possible to the really required
						startPoint = 0;
						numToReturn = nelements;
						status = setIntegerParam(hy8424Overflow_, 1);        		//overflow, samples collected is less than offset
					}
				}
				else
				{
					startPoint = pointer - (nelements - offset) * SAMPLEWORDS;		//total number greater than required
					numToReturn = nelements;
				}
			}
			break;
		case hy8424TRSimple:
			startPoint = 0;
			numToReturn = samples;
            break;
        default:
            status = asynError;
            break;
        }
		
		if(mode != hy8424TRVT)
		{
			for(i=0; i<(int)nelements; i++)
    	    {
    	        if(i < numToReturn)
    	        {
					if (mode == hy8424TRSimple)
					{
						*(data + i) = toVoltage((double)(this->convMemoryBase + channel * this->banksize)[startPoint]);
						startPoint++;
					}
					else
					{
						*(data + i) = toVoltage(((double)((epicsInt32)(this->convMemoryBase + channel * this->banksize)[startPoint] +
							(this->convMemoryBase + channel * this->banksize)[startPoint + 1] * 65536)) / average);
						startPoint += 2;
					}
    	        }
    	        else
    	            *(data + i) = 0;
    	        
    	        if(startPoint >= this->banksize)
    	        	startPoint = 0;                          							/* if pointer reaches the bottom of the memory, reset it */
    	    }
    	    if (nelements > (size_t)numToReturn)                                        /* if number to return less than number of elements, NORD = number of return */
    	        *nactual = numToReturn;
    	    else
    	        *nactual = nelements;                                               	/* if number to return is greater or equal number of elements, NORD = number of elements */
		}else																			// transient recorder voltage trigger mode
		{
			if((pointer < buffer) && (pointer > 0))
			{
				startPoint = pointer;
				numToReturn = buffer;
			}
			else if(pointer > buffer)
			{
				startPoint = 0;
				numToReturn = pointer;
			}
			else		//pointer == 0
			{
				startPoint = 0;
				numToReturn = buffer + samples;
			}
			
			for(i=0; i<(int)nelements; i++)
		    {
		        if (i < numToReturn)
		        {
	            	*(data + i) = toVoltage((double) (this->convMemoryBase + channel * this->banksize)[startPoint]);
		            startPoint++;
		        }
		        else
				{
					*(data + i) = 0;
					continue;
				}
		        
		        if((pointer < buffer) && (pointer > 0))
		        {
		        	if(startPoint >= numToReturn)
		        		startPoint = 0;                          						/* if pointer reaches the bottom of the memory, reset it */
		        }
		    }
	        *nactual = (int)nelements >= numToReturn ? numToReturn : nelements;              /* if number to return is greater or equal number of elements, NORD = number of elements */
		}
	}
    else if(function == hy8424SweepData_)
	{    	
    	pointer = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);
   	
		switch (mode)
		{
			case hy8424TRSimple:
				startPoint = this->adc->reg.tr[channel].ptrLow[0] + (this->adc->reg.tr[channel].ptrLow[1] << 16);
				uppointer = this->adc->reg.tr[channel].ptrHigh[0] + (this->adc->reg.tr[channel].ptrHigh[1] << 16);

				if((startPoint == 0) && (uppointer == 0))												// if both are 0, first try the ones from interrupt
				{
					startPoint = this->lastLowPointer;
					uppointer = this->lastHighPointer;
				}
				
				if((startPoint == 0) && (uppointer == 0))
				{
					for(i=0; i<(int)nelements; i++)
						*(data + i) = 0;																/* if all trails the pointers are still 0, nothing to present */
					*nactual = (int)nelements;
				}
				else
				{
					if(uppointer == 0)																	/* if upper pointer is 0, check if it is full */
					{
						if(((pointer+1) == 0) && (this->adc->reg.status & CSR_F))
							numToReturn = this->banksize;												/* if ADC memory is full, give it maximum */
						else
							numToReturn = pointer;														/* otherwise give it to the end */
					}else
						numToReturn = uppointer;														/* or to the pointer */

					for(i=0; i<(int)nelements; i++)
					{
						if (i < numToReturn)
						{
							*(data + i) = toVoltage((double) (this->convMemoryBase + channel * this->banksize)[startPoint]);
							startPoint++;
						}
						else
							*(data + i) = 0;
            
						if(startPoint >= this->banksize)
							startPoint = 0;                          										/* if pointer reaches the bottom of the memory, reset it */
					}
			        *nactual = (int)nelements >= numToReturn ? numToReturn : nelements;              /* if number to return is greater or equal number of elements, NORD = number of elements */
				}
				break;
			case hy8424TRVT:
				startPoint = this->adc->reg.tr[trchannel].ptrLow[0] + (this->adc->reg.tr[trchannel].ptrLow[1] << 16);
				uppointer = this->adc->reg.tr[trchannel].ptrHigh[0] + (this->adc->reg.tr[trchannel].ptrHigh[1] << 16);
				
				if((startPoint == 0) && (uppointer == 0))
				{
					for(i=0; i<(int)nelements; i++)
						*(data + i) = 0;
					*nactual = nelements;
				}
				else 
				{
					startPoint = buffer;
					if(uppointer > 0)
						numToReturn = uppointer - buffer;
					else
						numToReturn = samples;
		
					for(i=0; i<(int)nelements; i++)
		    	    {
        			    if (i < numToReturn)
		    	        {
       			        	*(data + i) = toVoltage((double) (this->convMemoryBase + channel * this->banksize)[startPoint]);
            	    		startPoint++;
		    	        }
        			    else
            	    		*(data + i) = 0;
            		}
			        *nactual = (int)nelements >= numToReturn ? numToReturn : nelements;              /* if number to return is greater or equal number of elements, NORD = number of elements */
				}
				break;
			default:
				break;
		}
	}

	status = asynSuccess;
	return( status );
}

/*********************************************************/
/*                                                      */
/* Private Sector                                      */
/*                                                    */
/*****************************************************/

void Hy8424Asyn::setupADC( int runtime)
{
	int mode, fastadc, enable, range;
	
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8424FastADC_, &fastadc);
	getIntegerParam(hy8424Range_, &range);
	getIntegerParam(hy8424SetMode_, &mode);

/* this first write to the CSR sets this bit and clears all others */
    this->adc->reg.status = 0;                                  			/* stop ADC at the very beginning */
	if((mode != hy8424TRSimple) && (mode != hy8424TRVT))					/* if it is transit recorder modes, fastadc param is ignored */
	{
	    if( fastadc )
    	    this->adc->reg.extStatus = 0;                                   /* normal mode doesn't need interrupt on each conversion */
	    else
    		this->adc->reg.extStatus = EXTCSR_ISE;                          /* clear extension CSR and set single conversion interrupt enable for mca and EPID record support */   
    }else
    	this->adc->reg.trstatus = 0;										// if it is transient recorder modes, clear its csr first

/* clear pointer to zero at start */
    this->adc->reg.pointer[1] = 0;
    this->adc->reg.pointer[0] = 0;

/* initialise average */
   	setAverage( runtime );	                             		/* check average setting overflow */

/* set up the requested internal clock rate */
	setClockRate( runtime );	          

/* set up the requested external / internal clock */
	setExtClock( runtime );	              

/* set offset to check if it exceeds BANKSIZE when added to samples */
    setOffset( runtime );

/* set up samples */
	setSamples( runtime );
	
/* the mode setup of the ADC depends on the mode that has been requsted. */
	setRange( range );

/* the mode setup of the ADC depends on the mode that has been requsted. */
	setMode( runtime );

/* disable the ADC at the beginning if it is not continuous mode and wait for the enable command */
	if(runtime)
		getIntegerParam(hy8424Enable_, &enable);
	else
	{
		if(mode == hy8424Continuous)
			enable = ENABLEADC;
		else
			enable = DISABLEADC;			
	}
    setEnableDisable( enable );
}

int Hy8424Asyn::setClockRate( int runtime  )
{
	int clockRate, extClock;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetClockRate_, &clockRate);
	getIntegerParam(hy8424SetExtClock_, &extClock);

	if (this->rebooting)  epicsThreadSuspendSelf();
	
	if(extClock == 1)				/* for external clock */
	{
		if(clockRate >= 17)
	        this->adc->reg.clockRate = 18;
		return(asynSuccess);
	}
	
	/* for internal clock */
    if ((clockRate < 0) || (clockRate > 18)) return( asynError );

    this->adc->reg.clockRate = clockRate;
    return(asynSuccess);
}

int Hy8424Asyn::setExtClock( int runtime )
{
	int clockRate, extClock, enable;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetClockRate_, &clockRate);
	getIntegerParam(hy8424SetExtClock_, &extClock);
	getIntegerParam(hy8424Enable_, &enable);

	if ( this->rebooting )  epicsThreadSuspendSelf();
       
    if( runtime )                                                   /* change samples at realtime */
    { 
        this->adc->reg.status &= ~CSR_ARM;                          /* disarm ADC first*/
        if (extClock)
	       this->adc->reg.status |= CSR_XC;
	    else
	       this->adc->reg.status &= ~CSR_XC;
        setEnableDisable(enable);          							/* start acquisition if it is enabled */
    }else
    {
        if (extClock)
	      this->adc->reg.status |= CSR_XC;
	    else
	      this->adc->reg.status &= ~CSR_XC;
    }    
    return( asynSuccess );
}

int Hy8424Asyn::softwareTrigger()
{
	int mode, enable;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424Enable_, &enable);
 
    if((mode != hy8424Triggered ) && (mode != hy8424TRSimple ) && (mode != hy8424TRVT )) 
        return asynError;                                           /* software trigger at realtime only applies to trigger mode */
    if(enable)                                           			/* only trigger when enabled */
    {
        this->adc->reg.status &= ~CSR_ST;                           /* clear software trigger latch */
        this->adc->reg.status |= CSR_ST;                            /* software trigger to kick off the acquisition */
        this->adc->reg.status &= ~CSR_ST;                           /* clear software latch to enable hardware trigger afterwards */
		setIntegerParam(hy8424TriggerStatus_, TRIGGERRED_COLLECTING);
  		callParamCallbacks();
    }
    return( asynSuccess );
}

int Hy8424Asyn::softwareStop()
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetMode_, &mode);
 
    if((mode == hy8424TRSimple ) || (mode == hy8424TRVT ))
   		this->adc->reg.trstatus |= TRCSR_SS; 	                    /* software stop */
	else
        return asynError;                                         	/* software stop only for transient recorder modes */

	callParamCallbacks();
    return( asynSuccess );
}

int Hy8424Asyn::setSamples(int runtime )
{
	int mode, average, samples, offset, enable, fastadc; 
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetSample_, &samples);
	getIntegerParam(hy8424SetAverage_, &average);
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetOffset_, &offset);
	getIntegerParam(hy8424Enable_, &enable);
	getIntegerParam(hy8424FastADC_, &fastadc);

    switch(mode)
    {
        case hy8424Continuous:
            break;
        case hy8424Triggered:
			if (runtime) this->adc->reg.status &= ~CSR_ARM;                  /* disarm ADC first*/
            if(samples > this->banksize) 
            {
                samples = this->banksize;
				setIntegerParam(hy8424SetSample_, samples);
				this->samples = samples;
                setIntegerParam(hy8424Overflow_, 1);            /* samples setting exceeds limit */
            } else
                setIntegerParam(hy8424Overflow_, 0);            /* samples setting within limit */
            this->adc->reg.numConversions[0] = (samples + offset) & 0xFFFF;   	/* put the new number of conversion in low */
            this->adc->reg.numConversions[1] = (samples + offset) >> 16;   		/* put the new number of conversion in high */

			if (runtime) setEnableDisable(enable);  						/* enable trigger if it is allowed */
            break;
        case hy8424Gated:                                       /* gated mode has nothing to do with the samples */
            break;
        case hy8424TRSimple:
        case hy8424TRVT:
            this->adc->reg.numConversions[0] = samples & 0xFFFF;	/* put the number of conversion in low */
            this->adc->reg.numConversions[1] = samples >> 16;		/* put the number of conversion in high */
            break;
        default:
            return asynError;
            break;
    }    

	callParamCallbacks();
    return( asynSuccess );
}


int Hy8424Asyn::setMode( int runtime )
{
	int mode, enable, fastadc;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424Enable_, &enable);
	getIntegerParam(hy8424FastADC_, &fastadc);
    
    if ((mode < hy8424Continuous) || (mode > hy8424TRVT)) return(asynError);

    if( runtime ) this->adc->reg.status &= ~CSR_ARM;		        /* stop 8424 totally */

    switch(mode)
    {
        case hy8424Continuous:
            break;
        case hy8424Triggered:                                       /* do nothing. The adc will be enabled by setEnable call */
			setIntegerParam(hy8424Enable_, DISABLEADC);
			this->enable = DISABLEADC;
			setIntegerParam(hy8424AutoReenable_, DISABLEADC);
            this->adc->reg.status |= CSR_EX;                        /* enable trigger */
            this->adc->reg.status |= CSR_EE;                        /* enable number of conversion interrupt */
            break;
        case hy8424Gated:
			setIntegerParam(hy8424Enable_, DISABLEADC);
			this->enable = DISABLEADC;
			setIntegerParam(hy8424AutoReenable_, DISABLEADC);
            this->adc->reg.extStatus |= EXTCSR_SMI;                 /* set software inhibit to preventing any writing to the memory */
            this->adc->reg.status |= CSR_ET;                        /* enable memory inhibit bit */
            this->adc->reg.status |= CSR_EII;                       /* enable inhibit interrupt */
            this->adc->reg.status |= CSR_ARM;	                    /* Arm the 8424 to kick off acquiring continuously */
            break;
        case hy8424TRSimple:
			setIntegerParam(hy8424Enable_, DISABLEADC);
			this->enable = DISABLEADC;
			setIntegerParam(hy8424AutoReenable_, DISABLEADC);
            this->adc->reg.status |= CSR_EX;                        /* enable trigger */
            this->adc->reg.status |= CSR_EE;                        /* enable number of conversion interrupt */
            this->adc->reg.trstatus |= TRCSR_TR;                    /* enable transit recorder mode */
            this->adc->reg.trstatus &= ~TRCSR_VT;                   /* disable transit recorder voltage trigger */

            break;
        case hy8424TRVT:
			setIntegerParam(hy8424Enable_, DISABLEADC);
			this->enable = DISABLEADC;
			setIntegerParam(hy8424AutoReenable_, DISABLEADC);
            this->adc->reg.status |= CSR_EX;                        /* enable trigger */
            this->adc->reg.status |= CSR_EE;                        /* enable number of conversion interrupt */
            this->adc->reg.trstatus |= TRCSR_TR;                    /* enable transit recorder mode */
            this->adc->reg.trstatus |= TRCSR_VT;                    /* enable transit recorder voltage trigger */
            break;
        default:
            return asynError;
            break;
    }

	callParamCallbacks();
    return( asynSuccess );
}

int Hy8424Asyn::setRange( int range )
{
	if (this->rebooting)  epicsThreadSuspendSelf();
 
	if(range)
	    this->adc->reg.extStatus |= EXTCSR_R;                         /* set +/-5V */
    else
		this->adc->reg.extStatus &= ~EXTCSR_R;                        /* set +/- 10V */
	setIntegerParam(hy8424Range_, range);
	callParamCallbacks();
    return( asynSuccess );
}

int Hy8424Asyn::setEnableDisable( int enable )
{
	int mode, fastadc;
	if (this->rebooting)  epicsThreadSuspendSelf();

	setIntegerParam(hy8424Enable_, enable);
	this->enable = enable;
	
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424FastADC_, &fastadc);
 
	if (enable)                                                     /* if enable */
    {
        switch (mode)
        {
            case hy8424Continuous:
                this->adc->reg.status |= CSR_ARM;	                /* Arm the 8424, enable acquiring */
                break;
            case hy8424Triggered:
            case hy8424TRSimple:
            case hy8424TRVT:
                this->adc->reg.pointer[1] = 0;                      /* zero the address pointer */
                this->adc->reg.pointer[0] = 0;                      /* zero the address pointer */
                this->adc->reg.status &= ~CSR_EX;                   /* disable trigger to clear any pending triggers */
                this->adc->reg.status |= CSR_EX;                    /* enable trigger */
                this->adc->reg.status |= CSR_ARM;	                /* Arm the 8424, enable acquiring */
                setIntegerParam(hy8424TriggerStatus_, WAITING_TRIGGER);
                break;
            case hy8424Gated:
                this->adc->reg.pointer[1] = 0;                      /* zero the address pointer */
                this->adc->reg.pointer[0] = 0;                      /* zero the address pointer */
                this->adc->reg.extStatus &= ~EXTCSR_SMI;            /* clear software iiinhibit to allow hardware inhibit */
                setIntegerParam(hy8424TriggerStatus_, WAITING_TRIGGER);
                break;
            default:
                return asynError;
                break;
        }
    }
	else                                                            /* if disable */
    {
        switch (mode)
        {
            case hy8424Continuous:                                  /* do nothing. wait for the interrupt to finish the current acquisition then stop */
            case hy8424Triggered:
            case hy8424TRSimple:
            case hy8424TRVT:
                this->adc->reg.status &= ~CSR_ARM;	                /* disable 8424 */
                break;
            case hy8424Gated:
                 this->adc->reg.extStatus |= EXTCSR_SMI;            /* set software inhibit to preventing any writing to the memory */
               break;
            default:
                return asynError;
                break;
        }
    }
	
//	callParamCallbacks();
	return( asynSuccess );
}

int Hy8424Asyn::setOffset( int runtime )
{
	int mode, samples, offset; 
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetSample_, &samples);
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetOffset_, &offset);

    switch (mode)
    { 
        case hy8424Continuous:
        	break;
        case hy8424Triggered:
			if (offset > samples)
            {
				offset = samples / 2;
				setIntegerParam(hy8424SetOffset_, offset);
				setIntegerParam(hy8424Overflow_, 1);      			/* samples setting exceeds limit */
            } else
                setIntegerParam(hy8424Overflow_, 0);            	/* samples setting within limit */
            break;
        case hy8424Gated:
            if (offset > this->banksize) 
				setIntegerParam(hy8424Overflow_, 1);      			/* samples setting exceeds limit */
            else
                setIntegerParam(hy8424Overflow_, 0);            	/* samples setting within limit */
            break;
        case hy8424TRSimple:										// only simple transient recoder mode needs this
			this->adc->reg.trdelay = offset;
            break;
        case hy8424TRVT:
            break;
        default:
            return asynError;
            break;
    }
  	callParamCallbacks();
    return( asynSuccess );
}

int Hy8424Asyn::setAverage( int runtime )
{
	int average;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424SetAverage_, &average);

	if((average < 0) || (average > 32767))
	{
		average = 1;
		setIntegerParam(hy8424SetAverage_, 1);      									/* force average to 1 if error */	
	    return( asynError );		
	}

	this->adc->reg.average = average | 0x8000;											//enable averaging
    return( asynSuccess );
}


int Hy8424Asyn::setTRLowThreshold( int runtime, int channel )
{
	int mode, threshold;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(channel, hy8424TRThresholdLow_, &threshold);
	getIntegerParam(hy8424SetMode_, &mode);

	if((mode == hy8424TRSimple) || (mode == hy8424TRVT))
	{
		this->adc->reg.tr[channel].thresholdLow = (USHORT)threshold;				/* set the threshold low word */
		return( asynSuccess );
	}else
	    return( asynError );		
}

int Hy8424Asyn::setTRHighThreshold( int runtime, int channel )
{
	int mode, threshold;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(channel, hy8424TRThresholdHigh_, &threshold);
	getIntegerParam(hy8424SetMode_, &mode);

	if((mode == hy8424TRSimple) || (mode == hy8424TRVT))
	{
		this->adc->reg.tr[channel].thresholdHigh = (USHORT)threshold;			/* set the threshold low word */
		return( asynSuccess );
	}else
	    return( asynError );		
}

int Hy8424Asyn::setTRChannel( int runtime )
{
	int mode, channel;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424TRChannel_, &channel);
	getIntegerParam(hy8424SetMode_, &mode);

	if((channel < 0) || (channel > 3)) return( asynError );

	if(mode == hy8424TRVT)																//only transient recorder voltage trigger mode need this
	{
		this->adc->reg.trstatus |= channel;												/* set the voltage trigger channel */
		return( asynSuccess );
	}else
	    return( asynError );		
}

int Hy8424Asyn::setTRBuffer( int runtime )
{
	int mode, buffer, samples;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424TRBuffer_, &buffer);
	getIntegerParam(hy8424SetMode_, &mode);
	getIntegerParam(hy8424SetSample_, &samples);

	if((buffer < 0) || (buffer > 65535)) return( asynError );

	if(mode == hy8424TRVT)																//only transient recorder voltage trigger mode need this
	{
		if(samples+buffer > BANKSIZE)
		{
			setIntegerParam(hy8424Overflow_, 1);      									/* samples setting exceeds limit */	
			samples = BANKSIZE - buffer;												// modify number of conversions
			setIntegerParam(hy8424SetSample_, samples);
			this->samples = samples;
			setSamples(runtime);
		}

		this->adc->reg.buffer = buffer;													/* set the circular buffer size */
		return( asynSuccess );
	}else
	    return( asynError );		
}

int Hy8424Asyn::setTROutEdge( int runtime )
{
	int mode, outedge;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424TROutputEdge_, &outedge);
	getIntegerParam(hy8424SetMode_, &mode);

	if((outedge < 0) || (outedge > 1)) return( asynError );

	if((mode == hy8424TRSimple)	|| (mode == hy8424TRVT))								//only transient recorder mode need this 
	{
		if(outedge == 0)
			this->adc->reg.trstatus &= ~TRCSR_SOL;										/* set the high to trigger */
		else
			this->adc->reg.trstatus |= TRCSR_SOL;										/* set the low to trigger */
		return( asynSuccess );
	}else
	    return( asynError );		
}

int Hy8424Asyn::setTRInEdge( int runtime )
{
	int mode, inedge;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8424TRInputEdge_, &inedge);
	getIntegerParam(hy8424SetMode_, &mode);

	if((inedge < 0) || (inedge > 1)) return( asynError );

	if((mode == hy8424TRSimple)	|| (mode == hy8424TRVT))								//only transient recorder mode need this 
	{
		if(inedge == 0)
			this->adc->reg.trstatus &= ~TRCSR_SIL;										/* set the high to stop */
		else
			this->adc->reg.trstatus |= TRCSR_SIL;										/* set the low to stop */
		return( asynSuccess );
	}else
	    return( asynError );		
}

void Hy8424Asyn::rebootCallback()
{
    this->adc->reg.status &= ~CSR_ARM;                              	/* Clear the ARM bit  */
    this->rebooting = 1;				                            	/* flag we are rebooting */
}

/*********************************************************/
/*                                                      */
/* Floating Voltage Convertion Routines                */
/*                                                    */
/*****************************************************/

/* this routine convert the raw adc value to voltage, -10V ~ +10V or -5V ~ +5V as per the range setting.  */
double Hy8424Asyn::toVoltage( double rawval )
{
	int range;
	
	getIntegerParam(hy8424Range_, &range);

	if(range == 0)														// The range is +/-10V
	{
		return (double)((double)10 * ((double) 2* rawval - 0xFFF1 - 0x00F) / ((double) 0xFFF1 - 0x00F));			
	}else																// The range is +/-5V
	{
		return (double)((double)5 * ((double) 2* rawval - 0xFFF1 - 0x00F) / ((double) 0xFFF1 - 0x00F));			
	}    
}

int Hy8424Asyn::calcThreshold( double value )
{
	int range;
	
	getIntegerParam(hy8424Range_, &range);

	if(range == 0)														// The range is +/-10V
	{
		return (int) (value * 65535 / 20 + 0x8000);			
	}else																// The range is +/-5V
	{
		return (int) (value * 65535 / 10 + 0x8000);			
	}    
}


double Hy8424Asyn::calcThresholdRaw( int value )
{
	int range;
	
	getIntegerParam(hy8424Range_, &range);

	if(range == 0)														// The range is +/-10V
	{
		return (double) (value - 0x8000) * 20 /65535;			
	}else																// The range is +/-5V
	{
		return (double) (value - 0x8000) * 10 /65535;			
	}    
}

/*********************************************************/
/*                                                      */
/* External Wrapped Routines      					   */
/*                                                    */
/*****************************************************/

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void rebootCallbackC(void *drvPvt)
{
    Hy8424Asyn *pPvt = (Hy8424Asyn *)drvPvt;
    pPvt->rebootCallback();
}

static void INTSRC(int cardNumber)
{
    Hy8424Asyn *pPvt = driverTable[cardNumber];
    pPvt->intSR();
}

static void intQueuedTaskC(void *drvPvt)
{
    Hy8424Asyn *pPvt = (Hy8424Asyn *)drvPvt;
    pPvt->intQueuedTask();	
}

static void intSweepQueuedTaskC(void *drvPvt)
{
    Hy8424Asyn *pPvt = (Hy8424Asyn *)drvPvt;
    pPvt->intSweepQueuedTask();		
}

}

/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task and external Routines      */
/*                                                    */
/*****************************************************/
void Hy8424Asyn::intSR()
{
    int data[2], flag = 1, sweepFlag = 1;
    int sweepData[MAX_ADC_CHANNELS + 2];
    int i = 0;
    USHORT regData;
	int queueCount;
	
    data[0] = this->adc->reg.pointer[0] + (this->adc->reg.pointer[1]<<16);          /* index 0 stores the address pointer */
	if((data[0] == 0) && (this->adc->reg.status & CSR_F))
		data[0] = this->banksize;
    sweepData[0] = data[0];															/* index 0 stores the address pointer */

	if ((this->adc->reg.extStatus & EXTCSR_IS) && (this->adc->reg.extStatus & EXTCSR_ISE)) 		/* if it is single conversion interrupt */
    {
    	this->adc->reg.extStatus |= EXTCSR_IS;                                   	/* write 1 to clear singel conversion interrupt */
        regData = this->adc->reg.extStatus;                                         /* to just ensure write successful */
       
        switch (this->mode)
        {
            case hy8424Continuous:
            case hy8424Triggered:
	        	for (i=0; i<MAX_ADC_CHANNELS; i++)                                	/* the rest of the elements store 8 channel raw data */
	        	{
    	        	sweepData[i + 2] = this->adc->reg.data[i];	        			
	        	}
                break;
            case hy8424Gated:
				if ((!(this->adc->reg.status & CSR_MIS)) 
                    && (!(this->adc->reg.extStatus & EXTCSR_SMI)))	                /* only collect data when gate is open */
				{
	        		for (i=0; i<MAX_ADC_CHANNELS; i++)                            	/* the rest of the elements store 8 channel raw data */
		        	{
	    	        	sweepData[i + 2] = this->adc->reg.data[i];	        			
		        	}
				} else
					sweepFlag = 0;			      					                /* if gate not opened, don't wake up the queue */
                break;
            default:
                sweepFlag = 0;
                break;
        }

    /* Wake up sweep queue task which calls callback routines if it is the correct interrupt. Note, the mode setting guarantees the interrupt settings */
        if (sweepFlag)
        {
			if(epicsMessageQueuePending(this->intSweepQId) >= (MAX_MESSAGES-1))
				messagesFailed++;
            else if (epicsMessageQueueTrySend(this->intSweepQId, sweepData, sizeof(sweepData)) == 0)
            	messagesSent++;
            else
            	messagesFailed++;
        }
    }
    else if(((this->adc->reg.status & CSR_MII) & (this->adc->reg.status & CSR_EII)) ||
    		(this->adc->reg.status & CSR_CC))										//only if there is an interreupt. conversion complete or gated interrupt
    {

    	/* re-enable interrupt */
        switch (this->mode)
        {
            case hy8424Triggered:
            case hy8424TRSimple:													
            case hy8424TRVT:														
            	this->adc->reg.status &= ~(CSR_CC | CSR_ARM);                   	/* clear interrupt, disarm ADC */
                break;
            case hy8424Gated:
            	this->adc->reg.status |= CSR_MII;                   				/* clear interrupt */
                break;
            default:
                flag = 0;
                break;
        }
        regData = this->adc->reg.status;                            				/* to just ensure write successful */

        /* Wake up queue task which calls callback routines if it is the correct interrupt. Note, the mode setting guarantees the interrupt settings */
        if (flag)
        {
			queueCount = epicsMessageQueuePending(this->intMsgQId);
			if(queueCount >= (MAX_MESSAGES-1))
				messagesFailed++;
            else if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
            	messagesSent++;
            else
            	messagesFailed++;
        }
    }
//    callParamCallbacks();
}

void Hy8424Asyn::intQueuedTask()
{
    int i;
    int data[2];
    int pointer, startPoint = 0, starttemp, uppointer = 0, numOfElement = 0;
	int mode, samples, average, offset, chan, reEnableState, buffer, trchannel;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intMsgQId, data, sizeof(data));
                 
    	getIntegerParam(hy8424SetMode_, &mode);
    	getIntegerParam(hy8424SetSample_, &samples);
    	getIntegerParam(hy8424SetAverage_, &average);
    	getIntegerParam(hy8424SetOffset_, &offset);
    	getIntegerParam(hy8424AutoReenable_, &reEnableState);
    	getIntegerParam(hy8424TRBuffer_, &buffer);
    	getIntegerParam(hy8424TRChannel_, &trchannel);    	               

       	/****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/
        pointer = data[0];
        this->lastPointer = data[0];                                                	// store pointer for periodic records (poll mode) 

        switch(mode)
        {
            case hy8424Triggered:
                if (pointer/SAMPLEWORDS >= samples) 
					numOfElement = samples;
                else
					numOfElement = pointer / SAMPLEWORDS;
				startPoint = 0;
                break;
            case hy8424Gated:
				if (offset < 0) offset = -offset;                                       //get correct offset */
                if (pointer/SAMPLEWORDS > offset)                                     		/* number of data collected is greater than offset */
                {
                    startPoint = offset*SAMPLEWORDS;
                    numOfElement = pointer/SAMPLEWORDS - offset;
                }else                                                           /* number of data collected is less than offset, no enough data, return all collected */
                {
                    startPoint = 0;
                    numOfElement = pointer/SAMPLEWORDS;
                }
                break;
		}

        starttemp = startPoint;
        for(chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
        	switch(mode)
        	{
			case hy8424TRSimple:
				startPoint = this->adc->reg.tr[chan].ptrLow[0] + (this->adc->reg.tr[chan].ptrLow[1] << 16);
				uppointer = this->adc->reg.tr[chan].ptrHigh[0] + (this->adc->reg.tr[chan].ptrHigh[1] << 16);

				this->lastLowPointer = startPoint;															/* for periodical records */
				this->lastHighPointer = uppointer;

				numOfElement = samples;
				for(i=0; i<(int)numOfElement; i++)
				{
           			rawData[chan][i] = (ULONG) (this->convMemoryBase + chan * this->banksize)[i];
               		adcData[chan][i] = toVoltage((double)rawData[chan][i]);
				}		
				break;
			case hy8424TRVT:
				uppointer = this->adc->reg.tr[trchannel].ptrHigh[0] + (this->adc->reg.tr[trchannel].ptrHigh[1] << 16);

				numOfElement = samples + buffer;											
				for(i=0; i<(int)numOfElement; i++)
				{
           			rawData[chan][i] = (ULONG) (this->convMemoryBase + chan * this->banksize)[i];
               		adcData[chan][i] = toVoltage((double)rawData[chan][i]);
				}
				break;
            case hy8424Triggered:
            case hy8424Gated:
            	// Pass int32Array interrupts 
                startPoint = starttemp;
                for(i=0; i<numOfElement; i++)
                {
           			rawData[chan][i] = ((ULONG) (this->convMemoryBase + chan * this->banksize)[startPoint] +
						(ULONG) (this->convMemoryBase + chan * this->banksize)[startPoint+1]*65536) / average;
               		adcData[chan][i] = toVoltage((double)rawData[chan][i]);
               		startPoint+=2;
                    if(startPoint >= this->banksize)                      	
                    {
                    	startPoint = 0;                    								// if pointer reaches the top of the memory, reset it 
                    }
                }
				break;      	
        	}

			doCallbacksInt32Array((epicsInt32 *) &rawData[chan], numOfElement, hy8424Data_, chan);
 			doCallbacksFloat64Array((epicsFloat64 *) &adcData[chan], numOfElement, hy8424FloatData_, chan);
 			
 			if(mode == hy8424TRVT)
 			{
 				doCallbacksInt32Array((epicsInt32 *) &rawData[chan][buffer], uppointer - buffer, hy8424SweepData_, chan);
 	 			doCallbacksFloat64Array((epicsFloat64 *) &adcData[chan][buffer], uppointer - buffer, hy8424FloatSweepData_, chan);
 			}else if(mode == hy8424TRSimple)
 			{
				if(uppointer == 0)														/* if upper pointer is 0, use the ones starting from lower threshold point */
					numOfElement = samples - startPoint;											
				else
					numOfElement = uppointer - startPoint;								/* from lower threshold to upper threshold */
 				doCallbacksInt32Array((epicsInt32 *) &rawData[chan][startPoint], numOfElement, hy8424SweepData_, chan);
 	 			doCallbacksFloat64Array((epicsFloat64 *) &adcData[chan][startPoint], numOfElement, hy8424FloatSweepData_, chan); 				
			}
        }

		this->adc->reg.pointer[0] = 0;											// reset pointer to 0
		this->adc->reg.pointer[1] = 0;

        /* re-enable interrupt */
        switch (mode)
        {
            case hy8424Continuous:                                                  	/* continuous mode re-initialise the adc in isr. Here nothing needs to be done */
                break;
            case hy8424Triggered:
            case hy8424Gated:
            case hy8424TRSimple:
            case hy8424TRVT:
                if (!reEnableState)                                           		/* if re-trigger is disabled, stop further triggers. To trigger again, need to do an enable command */
                    setEnableDisable( DISABLEADC );                       			/* disable trigger */
                else
                    setEnableDisable( ENABLEADC );                        			/* enable trigger */
                setIntegerParam(hy8424TriggerStatus_, FINISHED_AVAILABLE);        	/* finished and data is available */
                break;
            default:
                break;
        }
        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);
    }
    callParamCallbacks();
}


void Hy8424Asyn::intSweepQueuedTask()
{
    int sweepData[MAX_ADC_CHANNELS + 2];
    int chan;
    double dValue = 0.0;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intSweepQId, sweepData, sizeof(sweepData));

        /* Pass float64 interrupts */
        for (chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
            dValue = (double) sweepData[chan + 2];
    		setDoubleParam(chan, hy8424FloatSweepData_, dValue);
			callParamCallbacks(chan);
        }

        /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/

        /* Pass int32Array interrupts */
        for (chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
            chanData[chan] = (ULONG) sweepData[chan + 2];
        }
		doCallbacksInt32Array((epicsInt32 *) &chanData,
				MAX_ADC_CHANNELS, hy8424SweepData_, 0);

        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);
    }
}

